#include <linux/module.h>
#include <linux/kthread.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/err.h>
#include <linux/slab.h>
#include <linux/preempt.h>
#include <linux/string.h>

#define NR_OPTIONS	14

void get_start_time(struct timeval *tv)
{
	do_gettimeofday(tv);
}

void account_time(struct timeval *stv, struct timeval *etv)
{
	do_gettimeofday(etv);

	while (etv->tv_usec < stv->tv_usec) {
		etv->tv_sec--;
		etv->tv_usec += 1000000;
	}

	etv->tv_usec -= stv->tv_usec;
	etv->tv_sec -= stv->tv_sec;

	while (etv->tv_usec > 1000000) {
		etv->tv_usec -= 1000000;
		etv->tv_sec++;
	}

	printk("%lds %ldus", etv->tv_sec, etv->tv_usec);
}

extern void *__new_memcpy(void *to, void *from, size_t len);
extern void *__old_memcpy(void *to, void *from, size_t len);

char *str;

int init_module(void)
{
	struct timeval stv, etv;
	int loops, i, m;
	int len[NR_OPTIONS] = {
		1,
		7,
		8,
		16,
		32,
		48,
		64,
		80,
		128,
		256,
		512,
		1024,
		2048,
		4096,
	};
	unsigned long src_aligned, dst_aligned;

	str = (char *)__get_free_pages(GFP_KERNEL, 2);
	if (!str)
		return 0;

	loops = 5000000;

	printk("memcpy test (loops = %d):\n", loops);
	printk("Length\tSrc\tDest\told\t\tnew\n");
	printk("------\t---\t----\t---\t\t---\n");
	for (m = 0; m < NR_OPTIONS; m++) {
		for (src_aligned = 0; src_aligned < 8; src_aligned += 4) {
			for (dst_aligned = 0; dst_aligned < 8; dst_aligned += 4) {
				printk("%d\t%lu\t%lu\t",len[m], src_aligned, dst_aligned);
				yield();
				i = loops;
				get_start_time(&stv);

				local_irq_disable();
				while (i--)
					__old_memcpy(str + 2048 + dst_aligned, str + src_aligned, len[m]);
				local_irq_enable();

				account_time(&stv, &etv);

				printk("\t");
				yield();
				i = loops;
				get_start_time(&stv);

				local_irq_disable();
				while (i--)
					__new_memcpy(str + 2048 + dst_aligned, str + src_aligned, len[m]);
				local_irq_enable();

				account_time(&stv, &etv);
				printk("\n");
			}
		}
	}
	return 0;
}

void cleanup_module(void)
{
	free_pages((unsigned long)str, 2);
}

MODULE_LICENSE("GPL");
